#include <stdio.h>
#include <dos.h>
#include <math.h>
#include <conio.h>
#include <alloc.h>
#include <stdlib.h>

#define MAX(a,b) ((a > b) ? a : b)
#define MIN(a,b) ((a < b) ? a : b)

FILE *stl; // pointer to stl file
FILE *image; // pointer to bmp file

bool continuous = false; // used on debug
bool debug = false; // debug on/off
bool top = false; // start slicing from the top

char *dir = NULL; // directory output of bitmaps 
char *file_stl = NULL; // path to input stl file
char *buffer_line = NULL; // pointer to the buffer line of bitmap
char *cache_stl = NULL; // pointer to stl file in memory
char key = 0; // used on debug

float *p = NULL; // pointer temporary
float *cache_slice = NULL; // pointer to intersected lines of current slice
float *cache_scanline = NULL; // pointer to values in X of scanline-fill
float *vertices = NULL; // only used to be more simple the program
float minimum[3]; // minimum coordinated of model
float maximum[3]; // maximum coordinated of model
float _minimum[2]; // minimum coordinated of the generated slice
float _maximum[2]; // maximum coordinated of the generated slice
float dpix = 300; // resolution output in X of the image
float dpiy = 300; // resolution output in X of the image
float scale[3] = {1,1,1}; // default scale 1
float angle[3] = {0,0,0}; // used to rotate the model
float increment; // thickness of the slice

unsigned long *ntt; // pointer to the number of triangles
unsigned long x,i,j; // used in some loops
unsigned long size; // stl file size
unsigned long BiWidth; // width of bitmap in pixels
unsigned long BiHeight; // height of bitmap in pixels
unsigned long BiXPPMeter; // resolution in pixels per meter in X
unsigned long BiYPPMeter; // resolution in pixels per meter in Y
unsigned long BytesPerLine; // bytes per line of bitmap file
unsigned long lines; // number of lines not horizontal that was intersected
unsigned long lines_slf; // number of lines intersected in the scanline-fill
unsigned long slice = 0; // number of bitmap

unsigned char Pallet[8] = {255,255,255,0,0,0,0,0}; // colors of bitmap

double factor, diff; // used in some calculations
double z; // current position of slice
double y; // current position of scanline-fill

#include "stl.h"

void main(int argc, char *argv[])
{
   clrscr(); // clear screen

   Parse_Arguments(argc,argv);

   Open_STL_File();

   Scale_and_Rotate();

   Print_Header_Triangles();

   Get_ABS(); // get the max and min coordinates of the stl model

   if((buffer_line = (char *) malloc(BytesPerLine)) == NULL) // allocates memory for a line of bitmap
   {
      printf("\nInsufficient memory for the buffer... (%lu)", BytesPerLine);
      exit(6);
   }

   Print_ABS(); // print the max and min values of stl model

if(increment==0)
{
   Create_Menu();
}
   for((top) ? z = maximum[2] : z = minimum[2];
       (top) ? z > minimum[2] : z < maximum[2];
       (top) ? (z -= increment, slice++) : (z += increment, slice++)) // each loop is a layer
   {
      Is_Between(); // check which lines was intersected
      if(lines == 0) // if no lines intersected, go to the next slice
          continue;
      if(lines == 1) // if found only one line, something is wrong
      {
         printf("Error - Lines = 1 Layer %f", z);
         getch();
         continue; // go to the next slice
      }      

      Create_BMP(); // create the bmp
      Get_ABS_Slice(); // get the max(X,Y) and min(X,Y) of the slice generated

      for(y = _minimum[1]; y < _maximum[1]; y += ((25.4/dpiy))) // start of scanline-fill
      {
         lines_slf = 0; // initializes the variable
         cache_scanline = NULL; // to realloc work equal to malloc

         for(x = 0; x < lines; x++) // find wich lines is intersected and store the X coordinate
         {
            vertices = &cache_slice[x*4];
            if(MIN(vertices[1],vertices[3]) <= y && MAX(vertices[1],vertices[3]) > y)
            {
               diff = MAX(vertices[1],vertices[3]) - MIN(vertices[1],vertices[3]);
               factor = (y-MIN(vertices[1],vertices[3]))/diff;
               if(vertices[1] < vertices[3])
                  diff = (((vertices[2] - vertices[0]) * factor) + vertices[0]);
               else
                  diff = (((vertices[0] - vertices[2]) * factor) + vertices[2]);
               if((p = (float *) realloc (cache_scanline, ((lines_slf+1)*4))) == NULL)
               {
                  printf("Error (re)allocating memory");
                  exit(7);
               }
               cache_scanline = p;
               cache_scanline[lines_slf] = diff;
               lines_slf++;
               if(debug) printf("\n:%f %f",diff, y);
            }
         }
         // in the scanline-fill the numbers of lines must be even, if is odd you never 
		 // know where is inside and outside
         if(lines_slf % 2)
         {
            printf("\nError - Scanline-Fill  Odd = %lu %f", lines_slf, y);
            free(cache_scanline);
            getch();
            continue;
         }

         Order_SLF(); // with the lines intersected, reorder in ascending order

         Clean_Buffer();

         // write the horizontal line in the bitmap, separating all X intersected in pairs
         for(x = 0; x < (lines_slf/2); x++)
         {
            float xse[2];
            xse[0] = cache_scanline[x*2]; // x start
            xse[1] = cache_scanline[1+(x*2)]; // x end
			// trace a line between the x start to the x end on the bitmap using the current y
            BMP_Line(((xse[0]-minimum[0])/(25.4/dpix)),((xse[1]-minimum[0])/(25.4/dpix)));
         }

         // writes the entire line
         unsigned long ty = (((maximum[1]-minimum[1])-(y-minimum[1]))/(25.4/dpiy));
         unsigned long tempa = 62+(BytesPerLine*ty);
         fseek(image,tempa,SEEK_SET);
         fwrite(buffer_line,1,BytesPerLine,image);
         free(cache_scanline); // deallocate from memory the x values generated on scanline-fill
      } // end of scanline-fill
      fclose(image); // close the bmp file
      printf("\n%lu",slice); // print the number of last slice generated
   }
   free(cache_slice); // deallocate from memory the lines generated when sliced
   free(buffer_line); // deallocate from memory the buffer line used to generate the bitmap
   free(cache_stl); // deallocate from memory the stl file
} // end of program